from django import template
from administration.models import Grade, Person
import maths.Elo as elo_calc
from maths.models import Answer, Test
import surprise
import pandas
import json
import datetime
from pathlib import Path
import math
import surprise.accuracy

register = template.Library()


@register.simple_tag
def grade_check(grade, person):
    """
    Check if a person has permission to se the gradeDetailView of a grade.
    :param grade: The grade
    :param person: The person
    :return: Boolean value, true if the person has access.
    """
    if person.role == 4:
        return True
    elif person.role == 3:
        if grade.school.school_administrator == person:
            return True
    elif person.role == 2:
        for teacherGrade in person.grades.all():
            if grade == teacherGrade:
                return True
    return False


@register.simple_tag()
def group_check(group, person):
    if person.role == 4:
        return True
    elif person.role == 3:
        if group.grade.school.school_administrator == person:
            return True
    return False


@register.simple_tag()
def get_collaborative_filtering_estimates(user_id, tasks_to_consider):
    '''
    Input: user id and a list of tasks that should be considered
    Output: List of tasks sorted by highest estimate(best match), MAE and the difference between
            highest estimate average estimate

    The function gets a list of candidate tasks based on Elo rating. It uses pandas for structuring of the data and
    surprise for getting the collaborative filtering estimates.
    '''

    answers = Answer.objects.all()
    ratings_dict = {'itemID': [],
                    'userID': [],
                    'rating': []}
    for answer in answers:
        ratings_dict['itemID'].append(str(answer.test_id))
        ratings_dict['userID'].append(str(answer.user_id))
        ratings_dict['rating'].append(str(answer.correct))

    df = pandas.DataFrame(ratings_dict)
    reader = surprise.Reader(rating_scale=(0, 1))
    data = surprise.Dataset.load_from_df(df[['userID', 'itemID', 'rating']], reader)
    train_set = data.build_full_trainset()

    algo = surprise.KNNBasic()
    algo.train(train_set)

    estimates = 0
    tests_with_predictions = []
    predictions = []

    for test_id in tasks_to_consider:
        prediction = algo.predict(str(user_id), str(test_id), r_ui=0, verbose=True)
        predictions.append(prediction)
        tests_with_predictions.append([prediction[1], prediction[3], get_elo_diff(user_id, prediction[1])])
        estimates += prediction[3]

    mae = surprise.accuracy.mae(predictions, False)

    average_estimate = round(estimates/len(tests_with_predictions), 2)
    sorted_tests_with_predictions = sorted(tests_with_predictions, key=lambda x: (x[1], -x[2]), reverse=True)

    above_average = round(sorted_tests_with_predictions[0][1] - average_estimate, 2)
    return [sorted_tests_with_predictions, above_average, mae]


def get_uncompleted_tests(user_id):
    '''
    Input: user id
    Output: list of uncompleted tasks for the user
    '''

    tests_completed = []  # This holds all indexes of answers the user has given
    for answer in Answer.objects.filter(user=user_id):
        tests_completed.append(answer.test_id)

    tests_uncompleted = []
    all_tests = Test.objects.all()
    for test in all_tests:
        if test.id not in tests_completed:
            tests_uncompleted.append(test.id)
    return tests_uncompleted


@register.simple_tag
def calculate_new_elo(user_id):
    '''
    Input: user id
    Output: void

    Function for calculation of new Elo for the user and the task completed, and saves it to DB. Will run directly
    upon completion of every task.

    '''

    # Get user from DB
    user = Person.objects.get(id=user_id)
    user_elo = user.elo_rating

    # Get last answer for user from DB
    all_answers_for_user = Answer.objects.filter(user=user_id)
    if not all_answers_for_user:
        return json.dumps(True)

    last_answer = all_answers_for_user[len(all_answers_for_user) - 1]

    # Get task from DB
    test = Test.objects.get(id=last_answer.test.id)

    if len(all_answers_for_user) <= 0:
        return json.dumps(False)
    if last_answer.elo_has_been_calculated:
        return json.dumps(True)

    new_user_elo = elo_calc.calculate_new_elo_for_user(calculate_K_user(user_id),
                                                       user_elo, int(last_answer.correct),
                                                       elo_calc.calculate_expected_success_for_user(user_elo,
                                                       test.elo_rating))

    # Reversed last answer represents the result of the last task from the opponents(task) point of view
    # 1 = user failed, 0 = user succeeded

    reversed_last_answer = 1
    if int(last_answer.correct) == 1:
        reversed_last_answer = 0

    new_test_elo = elo_calc.calculate_new_elo_for_task(calculate_K_test(test),
                                                       test.elo_rating,
                                                       int(reversed_last_answer),
                                                       elo_calc.calculate_expected_success_for_task(test.elo_rating,
                                                       user_elo))
    test.elo_rating = new_test_elo
    test.save()

    if new_user_elo < 1200:
        user.elo_rating = 1200
    else:
        user.elo_rating = new_user_elo

    user.save()

    user_logger(user, test, last_answer.correct)
    test_logger(user, test, reversed_last_answer)
    all_answers_for_user[len(all_answers_for_user) - 1].elo_has_been_calculated = True
    all_answers_for_user[len(all_answers_for_user) - 1].save()


def calculate_K_user(user_id):
    a = 125
    b = 0.2

    n = len(Answer.objects.filter(user=user_id))
    return a / (1 + b * n)


def calculate_K_test(test_id):
    return 40


@register.simple_tag
def get_elo_candidate_tasks(user_id):
    '''
    Input: user id
    Output: a list of candidate tasks, sorted by absolute difference to target success rate(0.65)

    The function generates a list of tasks with the properties:
    - Not completed by the user
    - Taks that have a rating >= user Elo -25 og <= user Elo + 75
    '''

    # Get user rating from DB
    user_elo = Person.objects.get(id=user_id).elo_rating

    test_elo_ratings = []
    for test in Test.objects.all():
        if test.id in get_uncompleted_tests(user_id):
            if (test.elo_rating <= user_elo + 75 and test.elo_rating >= user_elo - 25):
                test_elo_ratings.append([test.id, test.elo_rating])

    # Calculate expected outcome for all current tasks
    top_candidates = []
    for test_elo in test_elo_ratings:
        top_candidates.append([test_elo[0], test_elo[1], get_elo_diff(user_id, test_elo[0])])

    # Sort absolute difference to target success rate
    sorted_probabilities = sorted(top_candidates, key=lambda x: x[2], reverse=False)

    # Removes negligible columns
    elo_candidates = []

    for probability in sorted_probabilities:
        elo_candidates.append(probability[0])

    if len(sorted_probabilities) > 0:
        return elo_candidates
    else:
        return []


def get_elo_diff(user_id, test_id):
    '''
    Input: user id, test id
    Output: absolute difference to target success rate

    Target success rate is set to .65 based on previous research
    '''

    target_success_rate = 0.65

    test_elo = Test.objects.get(id=test_id).elo_rating
    user_elo = Person.objects.get(id=user_id).elo_rating
    probability = elo_calc.calculate_expected_success_for_user(user_elo, test_elo)

    return abs(probability - target_success_rate)


@register.simple_tag()
def get_next_task(user_id):
    '''
    Input: user id
    Output: best task, based on Elo and collaborative filtering

    the function gets a list of Elo candidates, calls collaborative filtering function if the user is in the test
    group, and then returns the task id of the best match.
    '''

    user = Person.objects.get(id=user_id)

    if user.inControlGroup:
        return get_next_task_for_control_group_user(user)

    elo_candidates = get_elo_candidate_tasks(user_id)

    if len(elo_candidates) == 0:
        # No more tasks for the user, redirect to home page
        return -1

    collaborative_filtering_estimates = get_collaborative_filtering_estimates(user_id, elo_candidates)

    filtering_candidates = collaborative_filtering_estimates[0]
    above_average = collaborative_filtering_estimates[1]
    mae = round(collaborative_filtering_estimates[2], 4)

    log_rs_data_to_user_file(user, above_average, mae)
    return filtering_candidates[0][0]


def get_next_task_for_control_group_user(user):
    '''
    Input: user
    Output: task id of best match, based on Elo only

    Chooses the task with the success rate closest to the success target success rate
    '''

    uncompleted_tests = []

    for test in Test.objects.all():
            if test.id in get_uncompleted_tests(user.id) and test.elo_rating <= user.elo_rating + 75 and test.elo_rating >= user.elo_rating - 25:
                uncompleted_tests.append([test.id, get_elo_diff(user.id, test.id)])

    if len(uncompleted_tests) == 0:
        return -1

    uncompleted_tests_sorted = sorted(uncompleted_tests, key=lambda x: x[1], reverse=False)
    log_rs_data_to_user_file(user, '-', '-')

    return uncompleted_tests_sorted[0][0]


def log_rs_data_to_user_file(user, above_average, mae):
    file_path = 'log/users/' + str(user.id) + '_' + user.username + '.csv'

    if not Path(file_path).is_file():
        file = open(file_path, 'w+')
        if user.inControlGroup:
            file.write('USER ID: ' + str(user.id) + '\tTHIS USER IS IN THE CONROL GROUP!\n')
        else:
            file.write('USER ID: ' + str(user.id) + '\n')

        file.write('Test,Test-Elo,User-Elo,K-value,Success, Collab filtering, MAE, Time\n')
        file.close()

    file = open(file_path, 'a+')

    file.write(str(above_average) + ';' + str(mae))
    file.close()


def user_logger(user, test, answer):
    file_path = 'log/users/' + str(user.id) + '_' + user.username + '.csv'

    if not Path(file_path).is_file():
        file = open(file_path, 'w+')
        if user.inControlGroup:
            file.write('USER ID: ' + str(user.id) + '\tTHIS USER IS IN THE CONROL GROUP!\n')
        else:
            file.write('USER ID: ' + str(user.id) + '\n')

        file.write('Test,Test-Elo,User-Elo,K-value,Success, Collab filtering, MAE, Time\n')
        file.close()

    file = open(file_path, 'r')
    lines = file.readlines()
    last_line = lines[len(lines) - 1]
    temp = last_line.split(';')

    if temp:
        estimate = temp[0]
    else:
        estimate = '-'
    if len(temp)>1:
        mae = temp[1]
    else:
        mae = '-'

    file.seek(0)
    file.close()
    file = open(file_path, 'w')
    file.truncate()

    for line in lines[:len(lines) - 1]:
        file.write(line)

    file.write(str(test.id) + ',' + str(test.elo_rating) + ',' + str(user.elo_rating) + ',' + str(
        '{0:.2f}'.format(calculate_K_user(user.id))) + ',' + str(answer) + ',' + str(estimate) + ',' + str(mae) + ',' + str(
        datetime.datetime.now()) + '\n')

    file.close()


def test_logger(user, test, answer):
    file_path = 'log/tests/' + str(test.id) + '_' + test.task_collection.test_name + '.csv'

    if not Path(file_path).is_file():
        file = open(file_path, 'w+')

        file.write('TEST ID: ' + str(test.id) + '\n')

        file.write('User,User-Elo,Test-Elo,K-value,Success,Time\n')
        file.close()

    file = open(file_path, 'a+')
    file.write(str(user.id) + ',' + str(user.elo_rating) + ',' + str(test.elo_rating) + ',' + str(
        '{0:.2f}'.format(calculate_K_test(test.id))) + ',' + str(answer) + ',' + str(datetime.datetime.now()) + '\n')
    file.close()


@register.simple_tag()
def get_last_result_for_user(user_id):
    all_answers_for_user = Answer.objects.filter(user=user_id)
    if not len(all_answers_for_user) == 0:
        last_answer = all_answers_for_user[len(all_answers_for_user) - 1].correct

        if int(last_answer) == 1:
            return 'Last answer was CORRECT!'
        else:
            return 'Last answer was WRONG!'
    else:
        return 'User have not completed any tests yet...'


def get_active_students():
    '''
    Reurns the number of users that are students and have completed at least one task
    '''

    all_users = Person.objects.all()
    active_students = []

    for user in all_users:
        answer = Answer.objects.filter(user=user.id)
        if user.role == 1 and len(answer) > 0:
            active_students.append(user)

    return active_students


@register.simple_tag()
def get_all_tests():
    return Test.objects.all()


@register.simple_tag()
def get_test_name(test_id):
    return Test.objects.get(id=test_id).task_collection.test_name


@register.simple_tag()
def get_number_of_answers_for_test(test_id):
    answers = Answer.objects.filter(test=test_id)
    if len(answers) == 0:
        return 0
    else:
        return len(answers)


@register.simple_tag()
def get_correct_answers_for_test(test_id):
    answers = Answer.objects.filter(test=test_id)
    correct_answers = 0

    for answer in answers:
        if answer.correct == str(1):
            correct_answers += 1

    return correct_answers


@register.simple_tag()
def get_incorrect_answers_for_test(test_id):
    answers = Answer.objects.filter(test=test_id)
    incorrect_answers = 0

    for answer in answers:
        if answer.correct == str(0):
            incorrect_answers += 1

    return incorrect_answers


@register.simple_tag()
def get_average_user_elo():
    all_users = Person.objects.all()
    sum_elo = 0
    counter = 0

    for user in all_users:
        if user.role == 1 and not user.elo_rating == 1300:
            sum_elo += user.elo_rating
            counter += 1

    if counter == 0:
        return -1
    return round(sum_elo / counter)


@register.simple_tag()
def get_average_test_elo():
    all_tests = Test.objects.all()
    sum_elo = 0

    if all_tests == 0:
        return -1

    for test in all_tests:
        sum_elo += test.elo_rating

    return round(sum_elo / len(all_tests))


@register.simple_tag()
def get_number_of_tests_completed():
    return len(Answer.objects.all())


@register.simple_tag()
def average_number_of_completed_tests():

    if len(get_active_students()) == 0:
        return -1

    return round((len(Answer.objects.all()) / len(get_active_students())), 1)


@register.simple_tag()
def get_highest_elo_for_users():
    all_users = Person.objects.all()
    users_sorted = sorted(all_users, key=lambda x: x.elo_rating, reverse=True)
    return str(users_sorted[0].elo_rating) + ' - ' + str(users_sorted[0].username) + ', id: ' + str(users_sorted[0].id)


@register.simple_tag()
def get_lowest_elo_for_users():
    all_users = Person.objects.all()
    users_sorted = sorted(all_users, key=lambda x: x.elo_rating, reverse=False)
    return str(users_sorted[0].elo_rating) + ' - ' + str(users_sorted[0].username) + ', id: ' + str(users_sorted[0].id)


@register.simple_tag()
def get_highest_elo_for_tests():
    all_tests = Test.objects.all()
    tests_sorted = sorted(all_tests, key=lambda x: x.elo_rating, reverse=True)
    return str(tests_sorted[0].elo_rating) + ' - id: ' + str(tests_sorted[0].id)


@register.simple_tag()
def get_lowest_elo_for_tests():
    all_tests = Test.objects.all()
    tests_sorted = sorted(all_tests, key=lambda x: x.elo_rating, reverse=False)
    return str(tests_sorted[0].elo_rating) + ' - id: ' + str(tests_sorted[0].id)


@register.simple_tag()
def get_standard_deviation_for_users():
    all_users = Person.objects.all()
    average_elo = get_average_user_elo()
    sum_diff_squared = 0

    if len(get_active_students()) == 0:
        return -1

    for user in all_users:
        if user.role == 1 and not user.elo_rating == 1300:
            sum_diff_squared += math.pow(user.elo_rating - average_elo, 2)

    return round(math.sqrt(sum_diff_squared / len(get_active_students())), 2)


@register.simple_tag()
def get_standard_deviation_for_tests():
    all_tests = Test.objects.all()
    average_elo = get_average_test_elo()
    sum_diff_squared = 0

    if len(all_tests) == 0:
        return -1

    for test in all_tests:
        sum_diff_squared += math.pow(test.elo_rating - average_elo, 2)



    return round(math.sqrt(sum_diff_squared / len(all_tests)), 2)


@register.simple_tag()
def get_median_users():
    if len(get_active_students()) == 0:
        return -1
    users_sorted = sorted(get_active_students(), key=lambda x: x.elo_rating, reverse=True)
    mid_index = len(users_sorted) / 2

    if mid_index % 2 == 0:
        # Returns the average of the two mid indices
        return (users_sorted[int(math.floor(mid_index))].elo_rating + users_sorted[
            int(math.ceil(mid_index))].elo_rating) / 2
    else:
        return users_sorted[int(mid_index)].elo_rating


@register.simple_tag()
def get_median_tests():
    all_tests = Test.objects.all()
    tests_sorted = sorted(all_tests, key=lambda x: x.elo_rating, reverse=True)
    mid_index = len(tests_sorted) / 2

    if mid_index % 2 == 0:
        # Returns the average of the two mid indices
        return (tests_sorted[int(math.floor(mid_index))].elo_rating + tests_sorted[
            int(math.ceil(mid_index))].elo_rating) / 2
    else:
        return tests_sorted[int(mid_index)].elo_rating


@register.simple_tag()
def get_test_elo_distribution():
    all_tests = Test.objects.all()


    distribution = [[0, '<1200', 0], [1, '1200-1225', 0], [2, '1225-1250', 0], [3, '1250-1275', 0], [4, '1275-1300', 0],
                    [5, '1300-1325', 0], [6, '1325-1350', 0], [7, '1350-1375', 0], [8, '1375-1400', 0],
                   [9, '1400-1425', 0], [10, '1425-1450', 0], [11, '1450-1475', 0], [12, '1475-1500', 0],
                    [13, '1500-1525', 0], [14, '1525-1550', 0], [15, '1550-1575', 0], [16, '1575-1600', 0],
                    [17, '1600-1625', 0], [18, '1625-1650', 0], [19, '1650-1675', 0], [20, '1675-1700', 0],
                    [21, '1700-1725', 0], [22, '1725-1750', 0], [23, '1750-1775', 0], [24, '1775-1800', 0], [25, '>1800', 0]]


    roof = 1200

    for test in all_tests:
        if test.elo_rating < 1200:
            distribution[0][2] += 1

        if test.elo_rating > 1800:
            distribution[len(distribution)-1][2] += 1

        for d in distribution[:len(distribution)-1]:
            if test.elo_rating < roof and test.elo_rating > roof - 25:
                d[2] += 1
            roof += 25
        roof = 1200

    return distribution


@register.simple_tag()
def export_distribution_to_CSV():
    date = datetime.datetime.now()
    formatted_date = str(date)[:16].replace(' ', '_').replace('-', '_')
    file_path = 'log/test_distribution'  + '_' + str(formatted_date) + '.csv'
    file = open(file_path, 'w+')

    content = get_test_elo_distribution()
    for tuple in content:
        file.write(str(tuple[0]) + ',' + str(tuple[1]) + ',' + str(tuple[2]) + '\n')

    file.close()








